﻿EnableExplicit

#PLAYER_SPEED = 300
#PLAYER_TURN  = 0.1

Structure Vec3
  x.f
  y.f
  z.f
EndStructure

Structure structLevel
  body.i
  particle_fire.i
  particle_torch1.i
  particle_torch2.i
  particle_snow.i
  light_sun.i
  light_fire.i
  light_torch1.i
  light_torch2.i
EndStructure

Structure structPlayer
  body.i
  camera.i
  nodeMain.i
  nodeCamera.i
  nodeForward.i
  nodeStrafe.i
  Move.Vec3
  Rotate.Vec3
EndStructure

Define.b            FullScreen, DoLoop = #True
Define.f            TimeSinceLastFrame
Define.structLevel  Level
Define.structPlayer Player

Macro InitCheck(Status, ErrorMsg, EndProgram=#True)
  If Not Status
    CompilerIf #PB_Compiler_Debugger
      Debug "ERROR: " + ErrorMsg
    CompilerElse
      MessageRequester("ERROR", ErrorMsg)
    CompilerEndIf
    If EndProgram
      End
    EndIf
  EndIf
EndMacro

Declare Level_Create()
Declare Level_Update()
Declare Player_Create()
Declare Player_Update()
Declare StatDisplay()

If UCase(ProgramParameter()) = "FULLSCREEN"
  FullScreen = #True
EndIf

CompilerIf #PB_Compiler_Debugger
  InitCheck(InitEngine3D(#PB_Engine3D_EnableCG|#PB_Engine3D_DebugLog), "Can Not load 3D engine.")
CompilerElse
  InitCheck(InitEngine3D(#PB_Engine3D_EnableCG), "Can not load 3D engine.")
CompilerEndIf
InitCheck(InitSprite(), "Can not load display library.")
InitCheck(InitKeyboard(), "Can not load keyboard library.")
InitCheck(InitMouse(), "Can not load mouse library.")

If FullScreen
  InitCheck(ExamineDesktops(), "Can not find desktops.")
  InitCheck(OpenScreen(DesktopWidth(0), DesktopHeight(0), DesktopDepth(0), "Shader Test", #PB_Screen_NoSynchronization), "Can not open graphic screen.")
Else
  InitCheck(OpenWindow(0,0,0,1024,768,"Shader Test",#PB_Window_SystemMenu|#PB_Window_ScreenCentered), "Can not open OS window.")
  InitCheck(OpenWindowedScreen(WindowID(0), 0, 0, WindowWidth(0), WindowHeight(0), 0, 0, 0, #PB_Screen_WaitSynchronization), "Can not open graphic screen in a window.")
EndIf

CreateSprite(0, ScreenWidth(), ScreenHeight())
StartDrawing(SpriteOutput(0))
Box(0, 0, SpriteWidth(0), SpriteHeight(0), 0)
DrawText(2, SpriteHeight(0)-20, "Loading ...", RGB(128, 128, 128), 0)
StopDrawing()
DisplaySprite(0, 0, 0)
FlipBuffers()

Add3DArchive("textures", #PB_3DArchive_FileSystem)
Add3DArchive("textures/skybox", #PB_3DArchive_FileSystem)
Add3DArchive("textures/effects", #PB_3DArchive_FileSystem)
Add3DArchive("scripts", #PB_3DArchive_FileSystem)
Add3DArchive("programs", #PB_3DArchive_FileSystem)
Add3DArchive("models", #PB_3DArchive_FileSystem)
Parse3DScripts()

WorldShadows(#PB_Shadow_Additive, 200, RGB(48, 48, 48), 4096)

EnableWorldPhysics(#True)
EnableWorldCollisions(#True)

WorldDebug(#PB_World_DebugNone)

Level_Create()
Player_Create()
StatDisplay()

While Engine3DFrameRate(#PB_Engine3D_Current) < 30
  If Not FullScreen
    While WindowEvent() : Wend
  EndIf
  RenderWorld()
  DisplaySprite(0, 0, 0)
  FlipBuffers()
Wend

Engine3DFrameRate(#PB_Engine3D_Reset)
FreeSprite(0)

Repeat
  
  If Not FullScreen
    While WindowEvent() : Wend
  EndIf
  
  Level_Update()
  Player_Update()
  
  TimeSinceLastFrame = RenderWorld() / 1000
  
  StatDisplay()
  
  FlipBuffers()
 
Until DoLoop = #False

End

Macro GetPosition(Object, Position)
  If IsEntity(Object)
    Position\x = EntityX(Object)
    Position\y = EntityY(Object)
    Position\z = EntityZ(Object)
  ElseIf IsNode(Object)
    Position\x = NodeX(Object)
    Position\y = NodeY(Object)
    Position\z = NodeZ(Object)
  EndIf
EndMacro

Macro AddVec3(v1, v2, vResult)
  vResult\x = v1\x + v2\x
  vResult\y = v1\y + v2\y
  vResult\z = v1\z + v2\z
EndMacro

Macro SubVec3(V1, V2, vResult)
  vResult\x = V1\x - V2\x
  vResult\y = V1\y - V2\y
  vResult\z = V1\z - V2\z
EndMacro

Macro SwitchSign(Variable)
  If Variable >= 0
    Variable = -1
  Else
    Variable = 1
  EndIf
EndMacro

Procedure Level_Create()
  
  Shared      Level
  Protected.i tempMaterial, tempMesh
  
  tempMaterial = GetScriptMaterial(#PB_Any, "torchfire")
  
  SkyBox("night.png")
  
  With Level
    
    tempMesh    = LoadMesh(#PB_Any, "castle2.mesh")
    \body = CreateEntity(#PB_Any, MeshID(tempMesh), #PB_Material_None)
    EntityPhysicBody(\body, #PB_Entity_StaticBody)
    EntityRenderMode(\body, #PB_Entity_CastShadow)
    
    ; fireplace 0.4, 5.1, -51.3
    GetScriptParticleEmitter(0, "big_fire")
    MoveParticleEmitter(0, 0.04, 0.55, -5.5, #PB_Absolute)
    
    ; torch1 100.7, 15.8, 20.0
    ;GetScriptParticleEmitter(1, "small_fire")
    ;MoveParticleEmitter(1, 9.95, 4.0, 2.0, #PB_Absolute)
    CreateParticleEmitter(1, 0.1, 0.1, 0.1, #PB_Particle_Point, 9.95, 1.85, 2.0)
    ParticleMaterial(1, MaterialID(tempMaterial))
    ParticleSize(1, 0.2, 0.3)
    ParticleEmissionRate(1, 200)
    ParticleEmitterDirection(1, 0, 0.1, 0)
    ParticleTimeToLive(1, 0.1, 0.4)
    ParticleVelocity(1, 0.001, 0.005)
    ParticleColorRange(1, RGB(255, 0, 0), RGB(255, 64, 0))
    ParticleColorFader(1, 0.0, 1.0, 0.0)
    
    ; torch2 20.0, 15.8, 100.7
    CreateParticleEmitter(2, 0.1, 0.1, 0.1, #PB_Particle_Point, 2.0, 1.85, 9.95)
    ParticleMaterial(2, MaterialID(tempMaterial))
    ParticleSize(2, 0.2, 0.3)
    ParticleEmissionRate(2, 200)
    ParticleEmitterDirection(2, 0, 0.1, 0)
    ParticleTimeToLive(2, 0.1, 0.4)
    ParticleVelocity(2, 0.001, 0.005)
    ParticleColorRange(2, RGB(255, 128, 0), RGB(255, 64, 0))
    ParticleColorFader(2, 0.0, 1.0, 0.0)
    
    ; outside -7.2, 38.3, -255.9
    GetScriptParticleEmitter(3, "snow")
    MoveParticleEmitter(3, -0.75, 3.85, -25.0, #PB_Absolute)
    
;     \light_sun = CreateLight(#PB_Any, RGB(220, 220, 220), 0, 200, 100, #PB_Light_Point)
;     SetLightColor(\light_sun, #PB_Light_DiffuseColor, RGB(220, 220, 220))
;     SetLightColor(\light_sun, #PB_Light_SpecularColor, RGB(220, 220, 220))
;     LightAttenuation(\light_sun, 300 ,1)
    
    \light_fire = CreateLight(#PB_Any, RGB(240, 230, 140), 0.0, 1.01, -4.71, #PB_Light_Point)
    SetLightColor(\light_fire, #PB_Light_DiffuseColor, RGB(255, 215, 0))
    SetLightColor(\light_fire, #PB_Light_SpecularColor, RGB(240, 230, 140))
    LightAttenuation(\light_fire, 10, 1)
    
    \light_torch1 = CreateLight(#PB_Any, RGB(240, 230, 140), 9.5, 2.0, 2.0, #PB_Light_Point)
    SetLightColor(\light_torch1, #PB_Light_DiffuseColor, RGB(255, 215, 0))
    SetLightColor(\light_torch1, #PB_Light_SpecularColor, RGB(240, 230, 140))
    LightAttenuation(\light_torch1, 5, 1)
    
    \light_torch2 = CreateLight(#PB_Any, RGB(240, 230, 140), 2.0, 2.0, 9.5, #PB_Light_Point)
    SetLightColor(\light_torch2, #PB_Light_DiffuseColor, RGB(255, 215, 0))
    SetLightColor(\light_torch2, #PB_Light_SpecularColor, RGB(240, 230, 140))
    LightAttenuation(\light_torch2, 6, 1)
    
  EndWith
  
  AmbientColor(RGB(32, 32, 32))
  
  FreeMaterial(tempMaterial)
  FreeMesh(tempMesh)
  
EndProcedure

Procedure Level_Update()
  
  Shared      TimeSinceLastFrame, Level
  Static.b    torch1Dir, torch2Dir, fireDir
  Static.f    torch1Count, torch2Count, fireCount
  
  torch1Count + TimeSinceLastFrame
  If torch1Count >= Random(50, 5) / 1000
    SwitchSign(torch1Dir)
    MoveLight(Level\light_torch1, 0, torch1Count * torch1Dir, 0, #PB_Relative)
    LightAttenuation(Level\light_torch1, 5 - (10 * torch1Count), 1)
    torch1Count = 0
  EndIf
  
  torch2Count + TimeSinceLastFrame
  If torch2Count >= Random(100, 10) / 1000
    SwitchSign(torch2Dir)
    MoveLight(Level\light_torch2, 0, torch2Count * torch2Dir, 0, #PB_Relative)
    LightAttenuation(Level\light_torch2, 6 - (10 * torch2Count), 1)
    torch2Count = 0
  EndIf
  
   fireCount + TimeSinceLastFrame
   If fireCount >= Random(10, 1) / 1000
     SwitchSign(fireDir)
     MoveLight(Level\light_fire, fireCount * fireDir, 0, 0, #PB_Relative)
     LightAttenuation(Level\light_fire, 10 - (10 * fireCount), 1)
     fireCount = 0
   EndIf
  
EndProcedure

Procedure Player_Create()
  
  Shared      Player
  Protected.i tempMesh
  
  With Player
    
    \nodeMain    = CreateNode(#PB_Any,  0,   0,  0)
    \nodeCamera  = CreateNode(#PB_Any,  0,   0,  0)
    \nodeForward = CreateNode(#PB_Any,  0,   0, -1)
    \nodeStrafe  = CreateNode(#PB_Any, -1,   0,  0)
    
    tempMesh = CreateCylinder(#PB_Any, 0.4, 1.8)

    CreateEntity(\body, MeshID(tempMesh), #PB_Material_None, 0, 1, 0)
    EntityPhysicBody(\body, #PB_Entity_CapsuleBody, 0.8, 0.0, 0.0)
    HideEntity(\body, #True)
    
    \camera = CreateCamera(#PB_Any, 0, 0, 100, 100)
    CameraRange(\camera, 0.1, 100)
    
    AttachNodeObject(\nodeCamera, CameraID(\camera))
        
    AttachNodeObject(\nodeMain, NodeID(\nodeForward))
    AttachNodeObject(\nodeMain, NodeID(\nodeStrafe))
    AttachNodeObject(\nodeMain, NodeID(\nodeCamera))
    
  EndWith
  
  FreeMesh(tempMesh)
  
EndProcedure

Procedure Player_UpdateControls()
  
  Shared         Player, DoLoop, TimeSinceLastFrame
  Protected.f    speed, speed2
  Protected.Vec3 mainPos, forwardPos, strafePos, forwardDir, strafeDir, forwardMove, strafeMove
  
  With Player
    
    GetPosition(\nodeMain, mainPos)
    GetPosition(\nodeForward, forwardPos)
    GetPosition(\nodeStrafe, strafePos)
    SubVec3(forwardPos, mainPos, forwardDir)
    SubVec3(strafePos, mainPos, strafeDir)
    
  EndWith
  
  If ExamineKeyboard()
    
    If KeyboardReleased(#PB_Key_Escape)
      DoLoop = #False
    EndIf
    
    With Player\Keys
      
      speed  = TimeSinceLastFrame * #PLAYER_SPEED
      speed2 = speed * 0.5
      
      If KeyboardPushed(#PB_Key_W)
        forwardMove\x = forwardDir\x * speed
        forwardMove\z = forwardDir\z * speed
      ElseIf KeyboardPushed(#PB_Key_S)
        forwardMove\x = forwardDir\x * -speed2
        forwardMove\z = forwardDir\z * -speed2
      Else
        forwardMove\x = 0
        forwardMove\z = 0
      EndIf
      
      If KeyboardPushed(#PB_Key_A)
        strafeMove\x = strafeDir\x * speed2
        strafeMove\z = strafeDir\z * speed2
      ElseIf KeyboardPushed(#PB_Key_D)
        strafeMove\x = strafeDir\x * -speed2
        strafeMove\z = strafeDir\z * -speed2
      Else
        strafeMove\x = 0
        strafeMove\z = 0
      EndIf
      
    EndWith
    
  EndIf
  
  With Player
    
    AddVec3(forwardMove, strafeMove, \Move)
  
    If ExamineMouse()
      \Rotate\x = -MouseDeltaY() * #PLAYER_TURN
      \Rotate\y = -MouseDeltaX() * #PLAYER_TURN
    EndIf
  
  EndWith
  
EndProcedure

Procedure Player_Update()
  
  Shared Player
  
  Player_UpdateControls()
  
  With Player
    
    MoveEntity(\body, \Move\x, 0, \Move\z)
    RotateEntity(\body, 0, \Rotate\y, 0, #PB_Relative)
    If \Rotate\y = 0
      RotateEntity(\body, 0, EntityYaw(\body), 0, #PB_Absolute)
    EndIf
    
    RotateNode(\nodeCamera, \Rotate\x, 0, 0, #PB_Relative)
    If NodePitch(\nodeCamera) <= -60
      RotateNode(\nodeCamera, -60, 0, 0, #PB_Absolute)
    ElseIf NodePitch(\nodeCamera) >= 80
      RotateNode(\nodeCamera, 80, 0, 0, #PB_Absolute)
    EndIf
    
    MoveNode(\nodeMain, EntityX(\body), EntityY(\body) + 0.9, EntityZ(\body), #PB_Absolute)
    RotateNode(\nodeMain, 0, EntityYaw(\body), 0, #PB_Absolute) 
    
  EndWith

EndProcedure

Procedure StatDisplay()
  
  Shared   TimeSinceLastFrame
  Static.b init
  Static.l xPos, yPos, fCol, bCol
  Static.i statSprite, statFont
  Static.f counter
  Static.s txtCPU, txtTotalRAM, txtRam, txtCurFPS, txtMaxFPS, txtMinFPS, txtAvgFPS, txtTris
  
  If init = #False
    statsprite  = CreateSprite(#PB_Any, 280, 115, #PB_Sprite_AlphaBlending)
    statFont    = LoadFont(#PB_Any, "Courier New", 8, #PB_Font_HighQuality)
    xPos        = 0
    yPos        = ScreenHeight()-SpriteHeight(statSprite)
    bCol        = RGBA(0, 128, 0, 64)
    fCol        = RGBA(128, 128, 128, 196)
    txtCPU      = ReplaceString(CPUName(), Space(2), "")
    txtTotalRAM = Str(MemoryStatus(#PB_System_TotalPhysical) / 1000000)
    counter     = 100
    init        = #True
  EndIf
  
  counter + TimeSinceLastFrame
  If counter >= 1.0
    txtRAM    = "RAM:         " + Str(MemoryStatus(#PB_System_FreePhysical) / 1000000 ) + "/" + txtTotalRAM + " MB"
    txtCurFPS = "Current FPS: " + StrF(Engine3DFrameRate(#PB_Engine3D_Current), 1)
    txtMaxFPS = "Maximum FPS: " + StrF(Engine3DFrameRate(#PB_Engine3D_Maximum), 1)
    txtMinFPS = "Minimum FPS: " + StrF(Engine3DFrameRate(#PB_Engine3D_Minimum), 1)
    txtAvgFPS = "Average FPS: " + StrF(Engine3DFrameRate(#PB_Engine3D_Average), 1)
    txtTris   = "Triangles:   " + CountRenderedTriangles()
    StartDrawing(SpriteOutput(statSprite))
    DrawingMode(#PB_2DDrawing_AllChannels)
    DrawingFont(FontID(statFont))
    FrontColor(fCol)
    BackColor(bCol)
    Box(0, 0, SpriteWidth(statSprite), SpriteHeight(statSprite), bCol)
    DrawText(5, 5, txtCPU)
    DrawText(5, 20, txtRAM)
    DrawText(5, 35, txtCurFPS)
    DrawText(5, 50, txtAvgFPS)
    DrawText(5, 65, txtMinFPS)
    DrawText(5, 80, txtMaxFPS)
    DrawText(5, 95, txtTris)
    StopDrawing()
    counter = 0
  EndIf
  
  If IsSprite(statSprite)
    DisplayTransparentSprite(statSprite, xPos, yPos)
  EndIf
  
EndProcedure

; IDE Options = PureBasic 5.20 LTS (Windows - x86)
; CursorPosition = 65
; FirstLine = 15
; Folding = AA9
; EnableXP
; Executable = ShaderTest.exe
; CompileSourceDirectory
; EnablePurifier
; EnableCompileCount = 352
; EnableBuildCount = 5
; EnableExeConstant